/*
 *    apm.c
 *    Copyright (C) 2004  Thomas Nemeth <tnemeth@free.fr>
 *
 *    Based on work by Seiichi SATO <ssato@sh.rim.or.jp>
 *    Copyright (C) 2001,2002  Seiichi SATO <ssato@sh.rim.or.jp>
 *    and on work by Mark Staggs <me@markstaggs.net>
 *    Copyright (C) 2002  Mark Staggs <me@markstaggs.net>

 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.

 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "config.h"
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>


#ifdef linux
#include <sys/stat.h>
#endif


#if defined(solaris) || defined(sunos)
#include <unistd.h>
#include <fcntl.h>
#include <stropts.h>
#include <sys/battery.h>
#endif


#ifdef freebsd
#include <err.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <machine/apm_bios.h>
#endif


#include "apm.h"


#ifdef freebsd
# define APMDEV "/dev/apm"
#else
# ifdef linux
#  define APMDEV "/proc/apm"
# endif
#endif


#ifdef linux
# ifndef APM_32_BIT_SUPPORT
#  define APM_32_BIT_SUPPORT      0x0002
# endif
#endif


#ifdef freebsd
static int apm_read(apm_info_t temp_info);
#else
static int apm_read(ApmInfos *i);
#endif


void apm_init() {
    /* Check for APM support */
    if (!apm_exists()) {
#ifdef linux
        fprintf(stderr, "No APM support in kernel\n");
#else
        fprintf(stderr, "Unable to access APM info\n");
#endif
        exit(1);
    }
}


void apm_getinfos(ApmInfos *infos) {
#ifdef freebsd
    struct apm_info temp_info;
#endif

    if (
#if defined(linux) || defined(solaris) || defined(sunos)
    (apm_read(infos))
#else
# ifdef freebsd
    (apm_read(&temp_info))
# endif
#endif
     ) {
        fprintf(stderr, "Cannot read APM information.\n");
        exit(1);
    }
            
            
#ifdef freebsd     /* Convert status's */
    infos->ac_line_status     = (int)temp_info.ai_acline;
    infos->battery_status     = (int)temp_info.ai_batt_stat;
    /*infos->battery_percentage = (int)temp_info.ai_batt_life;*/
    /* on errors 255 is returned */
    if (255 == temp_info.ai_batt_life)
        infos->battery_percentage = -1;
    else
        infos->battery_percentage = (int)temp_info.ai_batt_life;
    infos->battery_time       = (int)temp_info.ai_batt_time;
#endif


#if defined(linux) || defined(freebsd)
    infos->battery_hoursleft =
        (infos->using_minutes) ? infos->battery_time / 60 : infos->battery_time;
#else
    infos->battery_hoursleft = infos->battery_time;
#endif
    infos->battery_minsleft = infos->battery_time % 60;
}


int apm_exists() {
    if (access(APMDEV, R_OK))
        return 0;
    else
        return 1;
}


#ifdef linux
static int apm_read(ApmInfos *i) {
    FILE        *str;
    char         units[10];
    char         buffer[100];
    int          retcode = 0;

    if (!(str = fopen(APMDEV, "r")))
        return 1;

    fgets(buffer, sizeof(buffer) - 1, str);
    buffer[sizeof(buffer) - 1] = '\0';
    sscanf(buffer, "%s %d.%d %x %x %x %x %d%% %d %s\n",
            (char *)i->driver_version,
            &i->apm_version_major,
            &i->apm_version_minor,
            &i->apm_flags,
            &i->ac_line_status,
            &i->battery_status,
            &i->battery_flags,
            &i->battery_percentage,
            &i->battery_time,
            units);
    i->using_minutes = !strncmp(units, "min", 3) ? 1 : 0;

    if (i->driver_version[0] == 'B') {
        strcpy((char *)i->driver_version, "pre-0.7");
        i->apm_version_major  = 0;
        i->apm_version_minor  = 0;
        i->apm_flags          = 0;
        i->ac_line_status     = 0xff;
        i->battery_status     = 0xff;
        i->battery_flags      = 0xff;
        i->battery_percentage = -1;
        i->battery_time       = -1;
        i->using_minutes      = 1;

        sscanf(buffer, "BIOS version: %d.%d", &i->apm_version_major, &i->apm_version_minor);
        fgets(buffer, sizeof(buffer) - 1, str);
        sscanf(buffer, "Flags: 0x%02x", &i->apm_flags);

        if (i->apm_flags & APM_32_BIT_SUPPORT) {
            fgets(buffer, sizeof(buffer) - 1, str);
            fgets(buffer, sizeof(buffer) - 1, str);

            if (buffer[0] != 'P') {

                if (!strncmp(buffer+4, "off line", 8))     i->ac_line_status = 0;
                else if (!strncmp(buffer+4, "on line", 7)) i->ac_line_status = 1;
                else if (!strncmp(buffer+4, "on back", 7)) i->ac_line_status = 2;

                fgets(buffer, sizeof(buffer) - 1, str);
                if (!strncmp(buffer+16, "high", 4))        i->battery_status = 0;
                else if (!strncmp(buffer+16, "low", 3))    i->battery_status = 1;
                else if (!strncmp(buffer+16, "crit", 4))   i->battery_status = 2;
                else if (!strncmp(buffer+16, "charg", 5))  i->battery_status = 3;

                fgets(buffer, sizeof(buffer) - 1, str);
                if (strncmp(buffer+14, "unknown", 7))      i->battery_percentage = atoi(buffer + 14);

                if (i->apm_version_major >= 1 && i->apm_version_minor >= 1) {
                    fgets(buffer, sizeof(buffer) - 1, str);
                    sscanf(buffer, "Battery flag: 0x%02x", &i->battery_flags);
                    fgets(buffer, sizeof(buffer) - 1, str);
                    if (strncmp(buffer+14, "unknown", 7))  i->battery_time = atoi(buffer + 14);
                }
            }
        }
    }

    if (i->battery_percentage > 100) i->battery_percentage = -1;
    fclose(str);

    return retcode;
}
#else
# ifdef freebsd
static int apm_read(apm_info_t temp_info) {
    int fd;
    if ( (fd = open(APMDEV, O_RDWR)) < 0) {
        return 1;
    } else if (ioctl(fd, APMIO_GETINFO, temp_info) == -1) {
        close(fd);
        return 1;
    } else {
        close(fd);
        return 0;
    }
}
# endif
#endif
#if defined(solaris) || defined(sunos)
static int apm_read(ApmInfos *i) {
    int fd;
    battery_t info;

    memset(i,0,sizeof(*i));
    if ((fd = open(APMDEV,O_RDONLY)) < 0) {
        perror("open");
        exit(1);
    }
    if (ioctl(fd,BATT_STATUS,&info) < 0) return 1;
    close(fd);

    i->battery_percentage = info.capacity;
    i->battery_time = info.discharge_time;
    i->using_minutes = 0;

    switch(info.status) {
        case EMPTY:             /* Battery has (effectively) no capacity */
            i->battery_status = 2;
            break;
        case LOW_CAPACITY:      /* Battery has less than 25% capacity */
            i->battery_status = 1;
            break;
        case MED_CAPACITY:      /* Battery has less than 50% capacity */
            i->battery_status = 1;
            break;
        case HIGH_CAPACITY:     /* Battery has less than 75% capacity */
        case FULL_CAPACITY:     /* Battery has more than 75% capacity */
            i->battery_status = 0;
            break;
        default:
            i->battery_status = 2;
            break;
    }

    switch(info.charge) {
        case DISCHARGE:         /* Battery is discharging (i.e. in use) */
            i->ac_line_status = 0;
            break;
        case FULL_CHARGE:       /* Battery is charging at its fastest rate */
        case TRICKLE_CHARGE:    /* Battery is charging at a slower rate */
        default:
            i->ac_line_status = 1;
            break;
    }

    if (i->battery_percentage > 100) i->battery_percentage = 100;

    return 0;
}
#endif

