/*
 * APM for root window.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <locale.h>
#include <string.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <librootapp/librootapp.h>
#include "config.h"
#include "apm.h"

#define EVENTMASK VisibilityChangeMask | ExposureMask


/* Global varriables */
static int end_program  = FALSE;

static char      *dpyname    = NULL;
static char      *fontname   = NULL;
static char      *geometry   = NULL;
static char      *mcolorname = NULL;
static char      *scolorname = NULL;
static ApmInfos   cur_apm_infos;
static unsigned long cm = 0, cs = 0;



/* Command line options and config file entries */
static Option opttab[] = {
    {'h', NULL,       NULL, O_HELP,   NULL,        FALSE, "display help"},
    {'v', NULL,       NULL, O_VERS,   NULL,        FALSE, "display version"},
    {'d', "display",  NULL, C_STRING, &dpyname,    FALSE, "display name"},
    {'c', "color",    NULL, C_STRING, &mcolorname, FALSE, "main color"},
    {'s', "color",    NULL, C_STRING, &scolorname, FALSE, "secondary color"},
    {'f', "font",     NULL, C_STRING, &fontname,   FALSE, "font to use"},
    {'g', "geom",     NULL, C_STRING, &geometry,   FALSE, "time geometry"},
    {0,   NULL,       NULL, C_NONE,   NULL,        FALSE, NULL}
};



/****** DISPLAYING FUNCTIONS ******/

void draw_online(RootApp *app, int h, int ch) {
    int ph = ch < 5 ? 5 : ch + (ch % 2);

    XDrawLine(display, app->bgpixmap, gc,
            1,                h + 1,       ph / 4 + 1,       h + 1);
    XDrawLine(display, app->bgpixmap, gc,
            1,                h + ph - 2,  ph / 4 + 1,       h + ph - 2);

    XFillRectangle(display, app->bgpixmap, gc,
            2 + ph / 4,       h,           ph - 1,           ph);

    XDrawLine(display, app->bgpixmap, gc,
            ph / 2 + ph,      h + 1,       ph / 2 + ph,      h + ph - 2);
    XDrawLine(display, app->bgpixmap, gc,
            1 + ph / 2 + ph,  h + 2,       7 + ph / 2 + ph,  h + 2);
}


void draw_charging(RootApp *app, int h) {
    XDrawLine(display, app->bgpixmap, gc,
            app->wsize.width / 2 - 3, h + 2, app->wsize.width / 2 + 3, h + 2);
    XDrawLine(display, app->bgpixmap, gc,
            app->wsize.width / 2 + 1, h,     app->wsize.width / 2 + 3, h + 2);
    XDrawLine(display, app->bgpixmap, gc,
            app->wsize.width / 2 + 1, h + 4, app->wsize.width / 2 + 3, h + 2);
}


void draw_battery(RootApp *app, int h, int ch) {
    XDrawRectangle(display, app->bgpixmap, gc,
            app->wsize.width - 12, h + 1, 8, ch);
    XFillRectangle(display, app->bgpixmap, gc,
            app->wsize.width - 4, h + 2, 2, ch - 1);
}


void draw_apm_status(RootApp *app, Bool ison) {
    char text[10];
    int  h = font_height();
    int  cw, ch = (h - 3) / 2;

    set_color(cm);
    bzero(text, 10);
    snprintf(text, 9, "%d:%02d",
            cur_apm_infos.battery_hoursleft,
            cur_apm_infos.battery_minsleft);
    root_print(app, text, app->wsize.width - 1 - text_width(text), 0);

    if (cur_apm_infos.battery_status == 3) {
        if (ison) draw_charging(app, h);
        draw_battery(app, h, ch);
    }

    if (cur_apm_infos.ac_line_status == 1) {
        draw_online(app, h, ch);
    } else {
        draw_battery(app, h, ch);
    }

    cw = cur_apm_infos.battery_percentage * (app->wsize.width - 2) / 100;
    XFillRectangle(display, app->bgpixmap, gc,
            1, app->wsize.height - 1 - ch, cw + 1, ch);

    set_color(cs);
    XDrawRectangle(display, app->bgpixmap, gc,
            0, app->wsize.height - 2 - ch, app->wsize.width - 1, ch + 1);
}


/****** PROCESSING FUNCTIONS ******/

void main_loop(RootApp *app) {
    XEvent   event;
    int      cont = TRUE;
    Bool     toggle = False;

    draw_apm_status(app, toggle);
    display_app(app);
    while (cont) {
        nextevent_or_timeout(&event, 1000);
        toggle = ! toggle;
        apm_getinfos(&cur_apm_infos);
        clear_bg(app);
        draw_apm_status(app, toggle);
        display_app(app);
        if (end_program) cont = FALSE;
    }
}


void set_app_size(RootApp *app) {
    int w = text_width("00:00:00");
    int h = 2 * (font_height() + 1);

    resize(app, geometry, 0, 0, w, h);
}


/****** SIGNALS ******/

void install_handler();


void signals_handler(int sig) {
    switch (sig) {
        case SIGINT:
        case SIGTERM:
        case SIGQUIT:
            end_program = TRUE;
            break;
    }
}


void install_handler() {
    signal(SIGTERM, signals_handler);
    signal(SIGQUIT, signals_handler);
    signal(SIGINT,  signals_handler);
}


void manage_signals() {
    struct sigaction sa;

    sa.sa_handler = SIG_IGN;
#ifdef SA_NOCLDWAIT
    sa.sa_flags = SA_NOCLDWAIT;
#else
    sa.sa_flags = 0;
#endif
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD, &sa, NULL);
    install_handler();
}


/****** MAIN ******/


int main(int argc, char *argv[]) {
    RootApp *app = NULL;

    /* General init */
    manage_signals();
    setlocale(LC_ALL, "");
    apm_init();
    init_rootapp(PACKAGE, PROGRAM, VERSION, AUTHORS);
    parse_argv(opttab, argc, argv);
    apm_getinfos(&cur_apm_infos);

    /* X11 init */
    open_x11(dpyname);
    if (fontname) load_font(fontname);
    if (mcolorname) cm = get_color(mcolorname);
    if (scolorname) cs = get_color(scolorname);

    /* Creation */
    app = create_app(EVENTMASK, geometry);
    set_app_size(app);

    main_loop(app);

    /* Clean */
    close_app(app);
    close_x11();
    free_vars(opttab);

    return 0;
}
